# ISO-Latin2 encoding (iso8859-2) for mkt1font and vpl2vpl

# This document is a modified version of Jan "Yenya" Kasprzak's
# mpage filter ISO-Latin.2. Yenya writes as follows:

#% Wed Dec  7 14:43:55 MET 1994
#% Written by Jan "Yenya" Kasprzak (kas@foresta.cz, kas@muni.cz) for mpage
#% filter. ISO Latin 2 (aka iso8859-2) is a character set, used in some
#% Middle- and East European countries (Czech republic, Slovakia, Poland,
#% Slovenia, Hungary, and some other -- I don't remember right now).
#%
#% You can redistribute this file under the terms of GNU General
#% Public License (GPL). If you have not your own copy of GPL,
#% mail me, I will send you one.
#%                                       -Yenya
#%
#%	This encoding vector describes real 8859-2 encoding, except of
#% some little differences: I was not able to find how to name characters,
#% which have accent shaped as right quote. Instead of these characters
#% I put there character with caron accent. If you find out how to name
#% characters with "right quote" accent, mail me. Or you can change it
#% yourself. You must change the following characters:
#% 
#% lcaron -> l with right quote,
#% Lcaron -> L  --''--
#% tcaron -> L  --''--
#% dcaron -> L  --''--
#% 
#% Tcaron and Dcaron are written with a _real_ caron.

127	.notdef
128	.notdef
129	.notdef
130	.notdef
131	.notdef
132	.notdef
133	.notdef
134	.notdef
135	.notdef
136	.notdef
137	.notdef
138	.notdef
139	.notdef
140	.notdef
141	.notdef
142	.notdef
143	.notdef
144	dotlessi
145	grave
146	acute
147	circumflex
148	tilde
149	.notdef
150	breve
151	.notdef
152	.notdef
153	.notdef
154	ring
155	.notdef
156	.notdef
157	hungarumlaut
158	ogonek
159	caron
160	.notdef
161	A ogonek
162	breve
163	Lslash
164	currency
165	L caron					# ? -- See above
166	S acute
167	section
168	dieresis
169	S caron
170	S cedilla
171	T caron
172	Z acute
173	hyphen
174	Z caron
175	Z dotaccent
176	degree
177	a ogonek
178	ogonek
179	lslash
180	acute
181	l caron					# ? -- See above
182	s acute
183	circumflex
184	cedilla
185	s caron
186	s cedilla
187	t caron					# ? -- See above
188	z acute
189	hungarumlaut
190	z caron
191	z dotaccent
192	R acute
193	A acute
194	A circumflex
195	A breve
196	A dieresis
197	L acute
198	C acute
199	C cedilla
200	C caron
201	E acute
202	E ogonek
203	E dieresis
204	E caron
205	I acute
206	I circumflex
207	D caron
208	Eth
209	N acute
210	N caron
211	O acute
212	O circumflex
213	O hungarumlaut
214	O dieresis
215	multiply
216	R caron
217	U ring
218	U acute
219	U hungarumlaut
220	U dieresis
221	Y acute
222	T cedilla
223	germandbls
224	r acute
225	a acute
226	a circumflex
227	a breve
228	a dieresis
229	l acute
230	c acute
231	c cedilla
232	c caron
233	e acute
234	e ogonek
235	e dieresis
236	e caron
237	i acute
238	i circumflex
239	d caron					# ? -- See above
240	eth
241	n acute
242	n caron
243	o acute
244	o circumflex
245	o hungarumlaut
246	o dieresis
247	divide
248	r caron
249	u ring
250	u acute
251	u hungarumlaut
252	u dieresis
253	y acute
254	t cedilla
255	dotaccent
