# CSX encoding for mkt1font and vpl2vpl
#
# Classical Sanskrit eXtended encoding for the representation of
# Indian languages in Roman script

128	C cedilla
129	u dieresis
130	e acute
131	a circumflex
132	a dieresis
133	a grave
134	a ring
135	c cedilla
136	e circumflex
137	e dieresis
138	e grave
139	i dieresis
140	i circumflex
141	i grave
142	A dieresis
143	A ring
144	E acute
145	ae
146	AE
147	o circumflex
148	o dieresis
149	o grave
150	u circumflex
151	u grave
152	y dieresis
153	O dieresis
154	U dieresis
155	cent
156	sterling
157	yen
158	.notdef
159	r underbar
160	a acute
161	i acute
162	o acute
163	u acute
164	n tilde
165	N tilde
166	l tilde
167	m overdot
168	amacron breve
169	imacron breve
170	umacron breve
171	amacron tilde
172	imacron tilde
173	n underbar
174	guillemotleft
175	guillemotright
176	.notdef
177	.notdef
178	.notdef
179	.notdef
180	.notdef
181	amacron acute
182	amacron grave
183	imacron acute
184	imacron grave
185	.notdef
186	.notdef
187	.notdef
188	.notdef
189	umacron acute
190	umacron grave
191	.notdef
192	.notdef
193	.notdef
194	.notdef
195	.notdef
196	.notdef
197	.notdef
198	runderdot acute
199	runderdot grave
200	.notdef
201	.notdef
202	.notdef
203	.notdef
204	.notdef
205	.notdef
206	.notdef
207	runderdotmacron acute
208	a tilde
209	i tilde
210	u tilde
211	e tilde
212	o tilde
213	e breve
214	o breve
215	l underbar
216	umacron tilde
217	.notdef
218	.notdef
219	.notdef
220	.notdef
221	.notdef
222	.notdef
223	.notdef
224	a macron
225	germandbls
226	A macron
227	i macron
228	I macron
229	u macron
230	U macron
231	r underdot
232	R underdot
233	runderdot macron
234	Runderdot macron
235	l underdot
236	L underdot
237	lunderdot macron
238	Lunderdot macron
239	n overdot
240	N overdot
241	t underdot
242	T underdot
243	d underdot
244	D underdot
245	n underdot
246	N underdot
247	s acute
248	S acute
249	s underdot
250	S underdot
251	.notdef
252	m underdot
253	M underdot
254	h underdot
255	H underdot
